/* evtgen compiler
   Copyright (C) 2003 Julien Olivain and LSV, CNRS UMR 8643 & ENS Cachan.

   This file is part of evtgen.

   evtgen is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   evtgen is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with evtgen; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * compiler.c
 * ast builder
 **/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "evtgen.h"

extern FILE *parserin;

static evtgen_t *evtgen_ctx_g;

static void final_pass(evtgen_t *ctx);

void
compile_config_file(evtgen_t *ctx)
{
  int ret;

  evtgen_ctx_g = ctx;

  if (ctx->config_file == NULL)
    ctx->config_file = strdup(DEFAULT_CONFIG_FILE);

  if (strcmp(ctx->config_file, "-"))
    parserin = xfopen(ctx->config_file, "r");

  ret = parserparse();
  if (ret > 0)
    {
      DPRINTF( ("Error while parsing %s\n", ctx->config_file) );
      exit(EXIT_FAILURE);
    }

  /* find and set prob tree root */
  ctx->prob_tree_root = strhash_get(ctx->randvar_hash, "main");
  if (ctx->prob_tree_root == NULL)
    {
      fprintf(stderr, "error: random variable @main not defined\n");
      exit(EXIT_FAILURE);
    }

  final_pass(ctx);
}

/* compute unallocated probs and resolve random variable references */
static void
final_pass(evtgen_t *ctx)
{
  random_var_t *v;
  value_t *e;
  text_token_t *t;
  float prob;

  for (v = ctx->randvar_list; v; v = v->next)
    {
      prob = (1.0f - v->allocated_prob) / (float) v->num_unallocated_prob;
      for (e = v->event_list; e; e = e->next)
        {
          if (e->prob == -1.0f)
            e->prob = prob;
          for (t = e->token_list; t; t = t->generic.next)
            {
              if (t->generic.type == TOKEN_VAR)
                {
                  t->randvar.var = strhash_get(ctx->randvar_hash,
                                               t->randvar.name);
                  if (t->randvar.var == NULL)
                    {
                      fprintf(stderr, "Undefined reference to @%s in declaration of @%s:%i.\n",
                              t->randvar.name, v->name, v->line);
                      exit(EXIT_FAILURE);
                    }
                }
            }
        }
    }
}

text_token_t *
build_string_token(char *str)
{
  text_token_t *tok;

  tok = xzmalloc( sizeof(text_token_t) );
  tok->string.type = TOKEN_STRING;
  tok->string.str = str;

  return (tok);
}

text_token_t *
build_builtin_var_token(char *name)
{
  text_token_t *tok;

  tok = xzmalloc( sizeof(text_token_t) );
  tok->builtin.type = TOKEN_BUILTIN_VAR;

  tok->builtin.var = strhash_get(evtgen_ctx_g->builtin_var_hash, name);
  if (tok->builtin.var == NULL)
    {
      fprintf(stderr, "builtin variable $%s doesn't exist\n", name);
      exit(EXIT_FAILURE);
    }

  return (tok);
}

text_token_t *
build_randvar_token(char *name)
{
  text_token_t *tok;

  tok = xzmalloc( sizeof(text_token_t) );
  tok->randvar.type = TOKEN_VAR;
  tok->randvar.name = name;

  return (tok);
}

value_t *
build_event(float prob, text_token_t *token_list)
{
  value_t *event;
  text_token_t *tail;
  text_token_t *elmt;
  text_token_t *next;

  event = xzmalloc(sizeof (value_t));
  event->prob = prob;
  /* we have to reverse the list to have the original order */
  elmt = token_list;
  tail = NULL;
  while (elmt)
    {
      next = elmt->generic.next;
      elmt->generic.next = tail;
      tail = elmt;
      elmt = next;
    }
  event->token_list = tail;

  return (event);
}

random_var_t *
build_randvar(char *name, int line, value_t *event_list)
{
  random_var_t *randvar;
  value_t *tail;
  value_t *elmt;
  value_t *next;
  float allocated_prob;
  int num_unallocated_prob;

  randvar = xzmalloc(sizeof(random_var_t));
  randvar->name = name;
  randvar->line = line;
  /* reverse list */
  elmt = event_list;
  tail = NULL;
  allocated_prob = 0.0f;
  num_unallocated_prob = 0;
  while (elmt)
    {
      if (elmt->prob < 0.0f)
        num_unallocated_prob++;
      else
        allocated_prob += elmt->prob;

      next = elmt->next;
      elmt->next = tail;
      tail = elmt;
      elmt = next;
    }
  randvar->event_list = tail;

  if (allocated_prob > 1.0f)
    {
      printf("error in randvar @%s:%i: total prob > 1.0 (%6.4f)\n", name, line, allocated_prob);
      exit(EXIT_FAILURE);
    }

  if ((num_unallocated_prob == 0) && (allocated_prob != 1.0f) )
    {
      printf("error in randvar @%s:%i: definition is not exhaustive (prob = %6.4f)\n", name, line, allocated_prob);
      exit(EXIT_FAILURE);
    }

  randvar->allocated_prob = allocated_prob;
  randvar->num_unallocated_prob = num_unallocated_prob;

  strhash_add(evtgen_ctx_g->randvar_hash, randvar, name);

  return (randvar);
}

void
build_randvar_list(random_var_t *randvar_list)
{
  random_var_t *tail;
  random_var_t *elmt;
  random_var_t *next;

  elmt = randvar_list;
  tail = NULL;
  while (elmt)
    {
      next = elmt->next;
      elmt->next = tail;
      tail = elmt;
      elmt = next;
    }
  evtgen_ctx_g->randvar_list = tail;
}
